@@ -62,8 +62,7 @@ public class BZip2CompressorInputStream extends CompressorInputStream implements
     private int nInUse;
 
     private InputStream in;
-
-    private int currentChar = -1;
+    private final boolean decompressConcatenated;
 
     private static final int EOF = 0;
     private static final int START_BLOCK_STATE = 1;
@@ -97,8 +96,9 @@ public class BZip2CompressorInputStream extends CompressorInputStream implements
     private BZip2CompressorInputStream.Data data;
 
     /**
-     * Constructs a new BZip2CompressorInputStream which decompresses bytes read from the
-     * specified stream.
+     * Constructs a new BZip2CompressorInputStream which decompresses bytes
+     * read from the specified stream. This doesn't suppprt decompressing
+     * concatenated .bz2 files.
      * 
      * @throws IOException
      *             if the stream content is malformed or an I/O error occurs.
@@ -106,20 +106,39 @@ public class BZip2CompressorInputStream extends CompressorInputStream implements
      *             if <tt>in == null</tt>
      */
     public BZip2CompressorInputStream(final InputStream in) throws IOException {
-        super();
-
-        this.in = in;
-        init();
+        this(in, false);
     }
 
-    /*
-     * (non-Javadoc)
-     * 
-     * @see java.io.InputStream#read()
+    /**
+     * Constructs a new BZip2CompressorInputStream which decompresses bytes
+     * read from the specified stream.
+     *
+     * @param in the InputStream from which this object should be created
+     * @param decompressConcatenated
+     *                     if true, decompress until the end of the input;
+     *                     if false, stop after the first .bz2 stream and
+     *                     leave the input position to point to the next
+     *                     byte after the .bz2 stream
+     *
+     * @throws IOException
+     *             if the stream content is malformed or an I/O error occurs.
+     * @throws NullPointerException
+     *             if <tt>in == null</tt>
      */
+    public BZip2CompressorInputStream(final InputStream in, final boolean decompressConcatenated) throws IOException {
+        this.in = in;
+        this.decompressConcatenated = decompressConcatenated;
+
+        init(true);
+        initBlock();
+    }
+
+    @Override
     public int read() throws IOException {
         if (this.in != null) {
-            return read0();
+            int r = read0();
+            count(r < 0 ? -1 : 1);
+            return r;
         } else {
             throw new IOException("stream closed");
         }
@@ -130,6 +149,7 @@ public class BZip2CompressorInputStream extends CompressorInputStream implements
      * 
      * @see java.io.InputStream#read(byte[], int, int)
      */
+    @Override
     public int read(final byte[] dest, final int offs, final int len)
         throws IOException {
         if (offs < 0) {
@@ -148,11 +168,14 @@ public class BZip2CompressorInputStream extends CompressorInputStream implements
 
         final int hi = offs + len;
         int destOffs = offs;
-        for (int b; (destOffs < hi) && ((b = read0()) >= 0);) {
+        int b;
+        while (destOffs < hi && ((b = read0()) >= 0)) {
             dest[destOffs++] = (byte) b;
+            count(1);
         }
 
-        return (destOffs == offs) ? -1 : (destOffs - offs);
+        int c = (destOffs == offs) ? -1 : (destOffs - offs);
+        return c;
     }
 
     private void makeMaps() {
@@ -162,103 +185,114 @@ public class BZip2CompressorInputStream extends CompressorInputStream implements
         int nInUseShadow = 0;
 
         for (int i = 0; i < 256; i++) {
-            if (inUse[i])
+            if (inUse[i]) {
                 seqToUnseq[nInUseShadow++] = (byte) i;
+            }
         }
 
         this.nInUse = nInUseShadow;
     }
 
     private int read0() throws IOException {
-        final int retChar = this.currentChar;
-
-        switch (this.currentState) {
+        switch (currentState) {
         case EOF:
             return -1;
 
         case START_BLOCK_STATE:
-            throw new IllegalStateException();
+            return setupBlock();
 
         case RAND_PART_A_STATE:
             throw new IllegalStateException();
 
         case RAND_PART_B_STATE:
-            setupRandPartB();
-            break;
+            return setupRandPartB();
 
         case RAND_PART_C_STATE:
-            setupRandPartC();
-            break;
+            return setupRandPartC();
 
         case NO_RAND_PART_A_STATE:
             throw new IllegalStateException();
 
         case NO_RAND_PART_B_STATE:
-            setupNoRandPartB();
-            break;
+            return setupNoRandPartB();
 
         case NO_RAND_PART_C_STATE:
-            setupNoRandPartC();
-            break;
+            return setupNoRandPartC();
 
         default:
             throw new IllegalStateException();
         }
-
-        return retChar;
     }
 
-    private void init() throws IOException {
+    private boolean init(boolean isFirstStream) throws IOException {
         if (null == in) {
             throw new IOException("No InputStream");
         }
-        if (in.available() == 0) {
-            throw new IOException("Empty InputStream");
+
+        int magic0 = this.in.read();
+        if (magic0 == -1 && !isFirstStream) {
+            return false;
+        }
+        int magic1 = this.in.read();
+        int magic2 = this.in.read();
+
+        if (magic0 != 'B' || magic1 != 'Z' || magic2 != 'h') {
+            throw new IOException(isFirstStream
+                    ? "Stream is not in the BZip2 format"
+                    : "Garbage after a valid BZip2 stream");
         }
-        checkMagicChar('B', "first");
-        checkMagicChar('Z', "second");
-        checkMagicChar('h', "third");
 
         int blockSize = this.in.read();
         if ((blockSize < '1') || (blockSize > '9')) {
-            throw new IOException("Stream is not BZip2 formatted: illegal "
-                                  + "blocksize " + (char) blockSize);
+            throw new IOException("BZip2 block size is invalid");
         }
 
         this.blockSize100k = blockSize - '0';
 
-        initBlock();
-        setupBlock();
-    }
+        this.bsLive = 0;
+        this.computedCombinedCRC = 0;
 
-    private void checkMagicChar(char expected, String position)
-        throws IOException {
-        int magic = this.in.read();
-        if (magic != expected) {
-            throw new IOException("Stream is not BZip2 formatted: expected '"
-                                  + expected + "' as " + position + " byte but got '"
-                                  + (char) magic + "'");
-        }
+        return true;
     }
 
     private void initBlock() throws IOException {
-        char magic0 = bsGetUByte();
-        char magic1 = bsGetUByte();
-        char magic2 = bsGetUByte();
-        char magic3 = bsGetUByte();
-        char magic4 = bsGetUByte();
-        char magic5 = bsGetUByte();
+        char magic0;
+        char magic1;
+        char magic2;
+        char magic3;
+        char magic4;
+        char magic5;
 
-        if (magic0 == 0x17 && magic1 == 0x72 && magic2 == 0x45
-            && magic3 == 0x38 && magic4 == 0x50 && magic5 == 0x90) {
-            complete(); // end of file
-        } else if (magic0 != 0x31 || // '1'
-                   magic1 != 0x41 || // ')'
-                   magic2 != 0x59 || // 'Y'
-                   magic3 != 0x26 || // '&'
-                   magic4 != 0x53 || // 'S'
-                   magic5 != 0x59 // 'Y'
-                   ) {
+        while (true) {
+            // Get the block magic bytes.
+            magic0 = bsGetUByte();
+            magic1 = bsGetUByte();
+            magic2 = bsGetUByte();
+            magic3 = bsGetUByte();
+            magic4 = bsGetUByte();
+            magic5 = bsGetUByte();
+
+            // If isn't end of stream magic, break out of the loop.
+            if (magic0 != 0x17 || magic1 != 0x72 || magic2 != 0x45
+                    || magic3 != 0x38 || magic4 != 0x50 || magic5 != 0x90) {
+                break;
+            }
+
+            // End of stream was reached. Check the combined CRC and
+            // advance to the next .bz2 stream if decoding concatenated
+            // streams.
+            if (complete()) {
+                return;
+            }
+        }
+
+        if (magic0 != 0x31 || // '1'
+            magic1 != 0x41 || // ')'
+            magic2 != 0x59 || // 'Y'
+            magic3 != 0x26 || // '&'
+            magic4 != 0x53 || // 'S'
+            magic5 != 0x59 // 'Y'
+            ) {
             this.currentState = EOF;
             throw new IOException("bad block header");
         } else {
@@ -300,7 +334,7 @@ public class BZip2CompressorInputStream extends CompressorInputStream implements
         this.computedCombinedCRC ^= this.computedBlockCRC;
     }
 
-    private void complete() throws IOException {
+    private boolean complete() throws IOException {
         this.storedCombinedCRC = bsGetInt();
         this.currentState = EOF;
         this.data = null;
@@ -308,8 +342,13 @@ public class BZip2CompressorInputStream extends CompressorInputStream implements
         if (this.storedCombinedCRC != this.computedCombinedCRC) {
             throw new IOException("BZip2 CRC error");
         }
+
+        // Look for the next .bz2 stream if decompressing
+        // concatenated files.
+        return !decompressConcatenated || !init(false);
     }
 
+    @Override
     public void close() throws IOException {
         InputStream inShadow = this.in;
         if (inShadow != null) {
@@ -751,9 +790,9 @@ public class BZip2CompressorInputStream extends CompressorInputStream implements
         return dataShadow.perm[zt][zvec - dataShadow.base[zt][zn]];
     }
 
-    private void setupBlock() throws IOException {
-        if (this.data == null) {
-            return;
+    private int setupBlock() throws IOException {
+        if (currentState == EOF || this.data == null) {
+            return -1;
         }
 
         final int[] cftab = this.data.cftab;
@@ -783,13 +822,12 @@ public class BZip2CompressorInputStream extends CompressorInputStream implements
         if (this.blockRandomised) {
             this.su_rNToGo = 0;
             this.su_rTPos = 0;
-            setupRandPartA();
-        } else {
-            setupNoRandPartA();
+            return setupRandPartA();
         }
+        return setupNoRandPartA();
     }
 
-    private void setupRandPartA() throws IOException {
+    private int setupRandPartA() throws IOException {
         if (this.su_i2 <= this.last) {
             this.su_chPrev = this.su_ch2;
             int su_ch2Shadow = this.data.ll8[this.su_tPos] & 0xff;
@@ -804,39 +842,39 @@ public class BZip2CompressorInputStream extends CompressorInputStream implements
             }
             this.su_ch2 = su_ch2Shadow ^= (this.su_rNToGo == 1) ? 1 : 0;
             this.su_i2++;
-            this.currentChar = su_ch2Shadow;
             this.currentState = RAND_PART_B_STATE;
             this.crc.updateCRC(su_ch2Shadow);
+            return su_ch2Shadow;
         } else {
             endBlock();
             initBlock();
-            setupBlock();
+            return setupBlock();
         }
     }
 
-    private void setupNoRandPartA() throws IOException {
+    private int setupNoRandPartA() throws IOException {
         if (this.su_i2 <= this.last) {
             this.su_chPrev = this.su_ch2;
             int su_ch2Shadow = this.data.ll8[this.su_tPos] & 0xff;
             this.su_ch2 = su_ch2Shadow;
             this.su_tPos = this.data.tt[this.su_tPos];
             this.su_i2++;
-            this.currentChar = su_ch2Shadow;
             this.currentState = NO_RAND_PART_B_STATE;
             this.crc.updateCRC(su_ch2Shadow);
+            return su_ch2Shadow;
         } else {
             this.currentState = NO_RAND_PART_A_STATE;
             endBlock();
             initBlock();
-            setupBlock();
+            return setupBlock();
         }
     }
 
-    private void setupRandPartB() throws IOException {
+    private int setupRandPartB() throws IOException {
         if (this.su_ch2 != this.su_chPrev) {
             this.currentState = RAND_PART_A_STATE;
             this.su_count = 1;
-            setupRandPartA();
+            return setupRandPartA();
         } else if (++this.su_count >= 4) {
             this.su_z = (char) (this.data.ll8[this.su_tPos] & 0xff);
             this.su_tPos = this.data.tt[this.su_tPos];
@@ -853,51 +891,51 @@ public class BZip2CompressorInputStream extends CompressorInputStream implements
             if (this.su_rNToGo == 1) {
                 this.su_z ^= 1;
             }
-            setupRandPartC();
+            return setupRandPartC();
         } else {
             this.currentState = RAND_PART_A_STATE;
-            setupRandPartA();
+            return setupRandPartA();
         }
     }
 
-    private void setupRandPartC() throws IOException {
+    private int setupRandPartC() throws IOException {
         if (this.su_j2 < this.su_z) {
-            this.currentChar = this.su_ch2;
             this.crc.updateCRC(this.su_ch2);
             this.su_j2++;
+            return this.su_ch2;
         } else {
             this.currentState = RAND_PART_A_STATE;
             this.su_i2++;
             this.su_count = 0;
-            setupRandPartA();
+            return setupRandPartA();
         }
     }
 
-    private void setupNoRandPartB() throws IOException {
+    private int setupNoRandPartB() throws IOException {
         if (this.su_ch2 != this.su_chPrev) {
             this.su_count = 1;
-            setupNoRandPartA();
+            return setupNoRandPartA();
         } else if (++this.su_count >= 4) {
             this.su_z = (char) (this.data.ll8[this.su_tPos] & 0xff);
             this.su_tPos = this.data.tt[this.su_tPos];
             this.su_j2 = 0;
-            setupNoRandPartC();
+            return setupNoRandPartC();
         } else {
-            setupNoRandPartA();
+            return setupNoRandPartA();
         }
     }
 
-    private void setupNoRandPartC() throws IOException {
+    private int setupNoRandPartC() throws IOException {
         if (this.su_j2 < this.su_z) {
             int su_ch2Shadow = this.su_ch2;
-            this.currentChar = su_ch2Shadow;
             this.crc.updateCRC(su_ch2Shadow);
             this.su_j2++;
             this.currentState = NO_RAND_PART_C_STATE;
+            return su_ch2Shadow;
         } else {
             this.su_i2++;
             this.su_count = 0;
-            setupNoRandPartA();
+            return setupNoRandPartA();
         }
     }
 
@@ -937,8 +975,6 @@ public class BZip2CompressorInputStream extends CompressorInputStream implements
         // ===============
 
         Data(int blockSize100k) {
-            super();
-
             this.ll8 = new byte[blockSize100k * BZip2Constants.BASEBLOCKSIZE];
         }
 
@@ -949,7 +985,7 @@ public class BZip2CompressorInputStream extends CompressorInputStream implements
          * I don't initialize it at construction time to avoid unneccessary
          * memory allocation when compressing small files.
          */
-        final int[] initTT(int length) {
+        int[] initTT(int length) {
             int[] ttShadow = this.tt;
 
             // tt.length should always be >= length, but theoretically
@@ -965,4 +1001,35 @@ public class BZip2CompressorInputStream extends CompressorInputStream implements
 
     }
 
+    /**
+     * Checks if the signature matches what is expected for a bzip2 file.
+     * 
+     * @param signature
+     *            the bytes to check
+     * @param length
+     *            the number of bytes to check
+     * @return true, if this stream is a bzip2 compressed stream, false otherwise
+     * 
+     * @since 1.1
+     */
+    public static boolean matches(byte[] signature, int length) {
+
+        if (length < 3) {
+            return false;
+        }
+
+        if (signature[0] != 'B') {
+            return false;
+        }
+
+        if (signature[1] != 'Z') {
+            return false;
+        }
+
+        if (signature[2] != 'h') {
+            return false;
+        }
+
+        return true;
+    }
 }
